/***************************************************************************

  vidhrdw.c

  Functions to emulate the video hardware of the machine.

Important!	There are two types of NeoGeo romdump - MVS & MGD2.  They are both
converted to a standard format in the vh_start routines.


Graphics information:

0x00000 - 0xdfff	: Blocks of sprite data, each 0x80 bytes:
	Each 0x80 block is made up of 0x20 double words, their format is:
	Word: Sprite number (16 bits)
	Byte: Palette number (8 bits)
	Byte: Bit 0: X flip
		  Bit 1: Y flip
		  Bit 2: Automatic animation flag (4 tiles?)
		  Bit 3: Automatic animation flag (8 tiles?)
		  Bit 4: MSB of sprite number (confirmed, Karnov_r, Mslug). See note.
		  Bit 5: MSB of sprite number (MSlug2)
		  Bit 6: MSB of sprite number (Kof97)
		  Bit 7: Unknown for now

	Each double word sprite is drawn directly underneath the previous one,
	based on the starting coordinates.

0x0e000 - 0x0ea00	: Front plane fix tiles (8*8), 2 bytes each

0x10000: Control for sprites banks, arranged in words

	Bit 0 to 3 - Y zoom LSB
	Bit 4 to 7 - Y zoom MSB (ie, 1 byte for Y zoom).
	Bit 8 to 11 - X zoom, 0xf is full size (no scale).
	Bit 12 to 15 - Unknown, probably unused

0x10400: Control for sprite banks, arranged in words

	Bit 0 to 5: Number of sprites in this bank (see note below).
	Bit 6 - If set, this bank is placed to right of previous bank (same Y-coord).
	Bit 7 to 15 - Y position for sprite bank.

0x10800: Control for sprite banks, arranged in words
	Bit 0 to 5: Unknown
	Bit 7 to 15 - X position for sprite bank.

Notes:

* If rom set has less than 0x10000 tiles then msb of tile must be ignored
(see Magician Lord).

***************************************************************************/

#include "driver.h"
#include "common.h"
#include "usrintrf.h"
//#include "vidhrdw/generic.h"


static data16_t *neogeo_vidram16;
static data16_t *neogeo_paletteram16;	/* pointer to 1 of the 2 palette banks */
static data16_t *neogeo_palettebank[2]; /* 0x100*16 2 byte palette entries */
static int neogeo_palette_index;
static data16_t neogeo_vidram16_modulo;
static data16_t neogeo_vidram16_offset;
static int high_tile;
static int vhigh_tile;
static int vvhigh_tile;
int no_of_tiles;
static int palette_swap_pending;
static int fix_bank;

extern data16_t *neogeo_ram16;
extern unsigned int neogeo_frame_counter;
extern int neogeo_game_fix;

static UINT32 *l_y_skip;
static UINT32 dda_y_skip[0x100][0x21][17];
static UINT8 yskip[0x100][0x21];
static UINT8 mmy[0x100][0x10];

extern int neogeo_machine_country;
int gururin_fix = 0;
int sdodgeb_fix = 0;

static int bitmap_pitch = 0;

/* SFMAME */

int sf_always;
int sf_raster;
int ng_tile_counter=0;
static char alt_bg=0,b_down=0;
static char sprlayer_on=1,t_down=0;
static char info_on=0,v_down=0;
static char fglayer_on=1,w_down=0;
static char sprites_on=1,a_down=0;
static char s_down=0,d_down=0;
static char f_down=0,g_down=0;
//static char sprite_xoffset=0,sprite_yoffset=0;
//static char sprite_xoffset_store=0,sprite_yoffset_store=0;
//static char slash_down=0,ast_down=0;
//static char minus_down=0,plus_down=0;
//static char sf_zoom=0;
//static char z_down=0,x_down=0;
static char paused=0,p_down=0;
static signed int tile_start=0;
static unsigned int tile_end=0x12c;
//static char sf_flag=0;

static void display_status(struct osd_bitmap *bitmap) {
    char buf[30];
    sprintf(buf,"%03d > %03d : %03d",tile_start,tile_end,ng_tile_counter);
    ui_text(bitmap,buf,Machine->uiwidth-strlen(buf)*Machine->uifontwidth,0);
}

static void get_sf_keys(void) {

    if (seq_pressed(input_port_type_seq(IPT_UI_PAUSE)) && !(paused && (keyboard_pressed(KEYCODE_LSHIFT) || (keyboard_pressed(KEYCODE_RSHIFT))))) {
        if (!p_down) {
            paused^=1;
            p_down=1;
        }
    }
    
    else {
        p_down=0;
    }

    if (sf_always || paused) {
        
/*        if (!sf_flag) {
            sprite_xoffset=sprite_xoffset_store;
            sprite_yoffset=sprite_yoffset_store;
            sf_flag=1;
        } */

        if (seq_pressed(input_port_type_seq(IPT_UI_RESET_SF))) {
            sprites_on=1;
            alt_bg=0;
            fglayer_on=1;
            sprlayer_on=1;
//            sprite_xoffset=0;
//            sprite_yoffset=0;
            info_on=0;
            sprites_on=1;
            tile_start=0;
            tile_end=0x12c;
        }

        if (seq_pressed(input_port_type_seq(IPT_UI_SPEEDUP))) {
            s_down=0;
            d_down=0;
            f_down=0;
            g_down=0;
        }

        if (seq_pressed(input_port_type_seq(IPT_UI_LAYER1))) {
            if (!w_down) {
                fglayer_on^=1;
                w_down=1;
            }
        }

        else {
            w_down=0;
        }

        if (seq_pressed(input_port_type_seq(IPT_UI_SPRITES))) {
            if (!t_down) {
                sprlayer_on^=1;
                t_down=1;
            }
        }

        else {
            t_down=0;
        }

        if (seq_pressed(input_port_type_seq(IPT_UI_SF_ON))) {
            if (!a_down) {
                if (!sprlayer_on) {
                    sprlayer_on=1;
                    sprites_on=0;
                    a_down=1;
                }
                
                else {
                    sprites_on^=1;
                    a_down=1;
                }
            }
        }

        else {
            a_down=0;
        }    
    
        if (seq_pressed(input_port_type_seq(IPT_UI_START_DEC)) && !sprites_on) {
            if (!s_down && tile_start > 0) {
                tile_start-=1;
                s_down=1;
                sprlayer_on=1;
            }
        }
    
        else {
            s_down=0;
        }
    
        if (seq_pressed(input_port_type_seq(IPT_UI_START_INC)) && !sprites_on) {
            if (!d_down && tile_start < tile_end) {
                tile_start+=1;
                d_down=1;
                sprlayer_on=1;
            }
            
            if (tile_start > tile_end) {
                tile_start=tile_end;
            }
        }
        
        else {
            d_down=0;
        }
    
        if (seq_pressed(input_port_type_seq(IPT_UI_END_DEC)) && !sprites_on) {
            if (!f_down && tile_end > tile_start) {
                tile_end-=1;
                f_down=1;
                sprlayer_on=1;
            }
                
            if (tile_end < tile_start) {
                tile_end=tile_start;
            }
        }
        
        else {
            f_down=0;
        }
    
        if (seq_pressed(input_port_type_seq(IPT_UI_END_INC)) && !sprites_on) {
            if (!g_down) {
                tile_end+=1;
                g_down=1;
                sprlayer_on=1;
            }
            
            if (tile_end > 0x17f) {
                tile_end=0x17f;
            }
        }
        
        else {
            g_down=0;
        }

/*      if (seq_pressed(input_port_type_seq(IPT_UI_SPRITE_HORIZ_DEC))) {
            if (!slash_down) {
                if (sprite_xoffset > -20) {
                    sprite_xoffset-=1;
                    slash_down=1;
                }
            }
        }
        
        else {
            slash_down=0;
        }

        if (seq_pressed(input_port_type_seq(IPT_UI_SPRITE_HORIZ_INC))) {
            if (!ast_down) {
                if (sprite_xoffset < 20) {
                    sprite_xoffset+=1;
                    ast_down=1;
                }
            }
        }
        
        else {
            ast_down=0;
        }

        if (seq_pressed(input_port_type_seq(IPT_UI_SPRITE_VERT_DEC))) {
            if (!plus_down) {
                if (sprite_yoffset > -20) {
                    sprite_yoffset-=1;
                    plus_down=1;
                }
            }
        }
        
        else {
            plus_down=0;
        }

        if (seq_pressed(input_port_type_seq(IPT_UI_SPRITE_VERT_INC))) {
            if (!minus_down) {
                if (sprite_yoffset < 20) {
                    sprite_yoffset+=1;
                    minus_down=1;
                }
            }
        }
        
        else {
            minus_down=0;
        }

        if (keyboard_pressed(KEYCODE_Z)) {
            if (!z_down) {
                if (sf_zoom < 20) {
                    sf_zoom+=1;
                    z_down=1;
                }
            }
        }
        
        else {
            z_down=0;
        }

        if (keyboard_pressed(KEYCODE_X)) {
            if (!x_down) {
                if (sf_zoom > -20) {
                    sf_zoom-=1;
                    x_down=1;
                }
            }
        }
        
        else {
            x_down=0;
        } */

        if (seq_pressed(input_port_type_seq(IPT_UI_INFO))) {
            if (!v_down) {
                info_on^=1;
                v_down=1;
            }
        }

        else {
            v_down=0;
        }
    
        if (seq_pressed(input_port_type_seq(IPT_UI_ALTBG))) {
            if (!b_down) {
                alt_bg^=1;
                b_down=1;
            }
        }
    
        else {
            b_down=0;
        }
    }
    
/*    else {
        if (sf_flag) {
            sprite_xoffset_store=sprite_xoffset;
            sprite_yoffset_store=sprite_yoffset;
            sprite_xoffset=0;
            sprite_yoffset=0;
            sf_flag=0;
        }
    } */
}

/******************************************************************************/
/* neo_draw_sprite                                                            */
/******************************************************************************/

#define MASK28  0xf0000000
#define MASK24  0x0f000000
#define MASK20  0x00f00000
#define MASK16  0x000f0000
#define MASK12  0x0000f000
#define MASK8   0x00000f00
#define MASK4   0x000000f0
#define MASK0   0x0000000f


#define INCREMENT_LINE() \
sprite += l_y_skip[l++] * dy;

#define LOW_TILE() \
tile = *sprite;

#define HIGH_TILE() \
tile = *(sprite + 1);

#define ADD_PITCH() \
bitmap += bitmap_pitch;

#define DRAW_PIXEL(x, shift) \
pixel = (tile & MASK##shift); if (pixel) bitmap[x] = paldata[pixel >> shift];

/******************************************************************************/
/* draw sprite no flipx                                                       */
/******************************************************************************/

static void draw_sprite8_16(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		DRAW_PIXEL( 1,  4)
		DRAW_PIXEL( 2,  8)
		DRAW_PIXEL( 3, 12)
		DRAW_PIXEL( 4, 16)
		DRAW_PIXEL( 5, 20)
		DRAW_PIXEL( 6, 24)
		DRAW_PIXEL( 7, 28)
		HIGH_TILE()
		DRAW_PIXEL( 8,  0)
		DRAW_PIXEL( 9,  4)
		DRAW_PIXEL(10,  8)
		DRAW_PIXEL(11, 12)
		DRAW_PIXEL(12, 16)
		DRAW_PIXEL(13, 20)
		DRAW_PIXEL(14, 24)
		DRAW_PIXEL(15, 28)
		ADD_PITCH()
	}
}


static void draw_sprite8_15(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		DRAW_PIXEL( 1,  4)
		DRAW_PIXEL( 2,  8)
		DRAW_PIXEL( 3, 12)
		DRAW_PIXEL( 4, 16)
		DRAW_PIXEL( 5, 20)
		DRAW_PIXEL( 6, 24)
		HIGH_TILE()
		DRAW_PIXEL( 7,  0)
		DRAW_PIXEL( 8,  4)
		DRAW_PIXEL( 9,  8)
		DRAW_PIXEL(10, 12)
		DRAW_PIXEL(11, 16)
		DRAW_PIXEL(12, 20)
		DRAW_PIXEL(13, 24)
		DRAW_PIXEL(14, 28)
		ADD_PITCH()
	}
}


static void draw_sprite8_14(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		DRAW_PIXEL( 1,  4)
		DRAW_PIXEL( 2,  8)
		DRAW_PIXEL( 3, 12)
		DRAW_PIXEL( 4, 16)
		DRAW_PIXEL( 5, 20)
		DRAW_PIXEL( 6, 24)
		HIGH_TILE()
		DRAW_PIXEL( 7,  0)
		DRAW_PIXEL( 8,  4)
		DRAW_PIXEL( 9,  8)
		DRAW_PIXEL(10, 12)
		DRAW_PIXEL(11, 16)
		DRAW_PIXEL(12, 20)
		DRAW_PIXEL(13, 24)
		ADD_PITCH()
	}
}


static void draw_sprite8_13(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		DRAW_PIXEL( 1,  4)
		DRAW_PIXEL( 2, 12)
		DRAW_PIXEL( 3, 16)
		DRAW_PIXEL( 4, 20)
		DRAW_PIXEL( 5, 24)
		HIGH_TILE()
		DRAW_PIXEL( 6,  0)
		DRAW_PIXEL( 7,  4)
		DRAW_PIXEL( 8,  8)
		DRAW_PIXEL( 9, 12)
		DRAW_PIXEL(10, 20)
		DRAW_PIXEL(11, 24)
		DRAW_PIXEL(12, 28)
		ADD_PITCH()
	}
}


static void draw_sprite8_12(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		DRAW_PIXEL( 1,  4)
		DRAW_PIXEL( 2,  8)
		DRAW_PIXEL( 3, 16)
		DRAW_PIXEL( 4, 20)
		DRAW_PIXEL( 5, 24)
		HIGH_TILE()
		DRAW_PIXEL( 6,  0)
		DRAW_PIXEL( 7,  4)
		DRAW_PIXEL( 8,  8)
		DRAW_PIXEL( 9, 16)
		DRAW_PIXEL(10, 20)
		DRAW_PIXEL(11, 24)
		ADD_PITCH()
	}
}


static void draw_sprite8_11(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  4)
		DRAW_PIXEL( 1,  8)
		DRAW_PIXEL( 2, 16)
		DRAW_PIXEL( 3, 20)
		DRAW_PIXEL( 4, 28)
		HIGH_TILE()
		DRAW_PIXEL( 5,  0)
		DRAW_PIXEL( 6,  4)
		DRAW_PIXEL( 7, 12)
		DRAW_PIXEL( 8, 16)
		DRAW_PIXEL( 9, 24)
		DRAW_PIXEL(10, 28)
		ADD_PITCH()
	}
}


static void draw_sprite8_10(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		DRAW_PIXEL( 1,  8)
		DRAW_PIXEL( 2, 16)
		DRAW_PIXEL( 3, 24)
		DRAW_PIXEL( 4, 28)
		HIGH_TILE()
		DRAW_PIXEL( 5,  0)
		DRAW_PIXEL( 6,  8)
		DRAW_PIXEL( 7, 16)
		DRAW_PIXEL( 8, 24)
		DRAW_PIXEL( 9, 28)
		ADD_PITCH()
	}
}


static void draw_sprite8_9(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		DRAW_PIXEL( 1,  8)
		DRAW_PIXEL( 2, 16)
		DRAW_PIXEL( 3, 24)
		HIGH_TILE()
		DRAW_PIXEL( 4,  0)
		DRAW_PIXEL( 5,  8)
		DRAW_PIXEL( 6, 16)
		DRAW_PIXEL( 7, 24)
		DRAW_PIXEL( 8, 28)
		ADD_PITCH()
	}
}


static void draw_sprite8_8(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		DRAW_PIXEL( 1,  8)
		DRAW_PIXEL( 2, 16)
		DRAW_PIXEL( 3, 24)
		HIGH_TILE()
		DRAW_PIXEL( 4,  0)
		DRAW_PIXEL( 5,  8)
		DRAW_PIXEL( 6, 16)
		DRAW_PIXEL( 7, 24)
		ADD_PITCH()
	}
}


static void draw_sprite8_7(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  4)
		DRAW_PIXEL( 1, 12)
		DRAW_PIXEL( 2, 20)
		HIGH_TILE()
		DRAW_PIXEL( 3,  0)
		DRAW_PIXEL( 4,  8)
		DRAW_PIXEL( 5, 16)
		DRAW_PIXEL( 6, 24)
		ADD_PITCH()
	}
}


static void draw_sprite8_6(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  4)
		DRAW_PIXEL( 1, 16)
		DRAW_PIXEL( 2, 28)
		HIGH_TILE()
		DRAW_PIXEL( 3,  0)
		DRAW_PIXEL( 4, 12)
		DRAW_PIXEL( 5, 24)
		ADD_PITCH()
	}
}


static void draw_sprite8_5(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  8)
		DRAW_PIXEL( 1, 20)
		HIGH_TILE()
		DRAW_PIXEL( 2,  0)
		DRAW_PIXEL( 3, 12)
		DRAW_PIXEL( 4, 24)
		ADD_PITCH()
	}
}


static void draw_sprite8_4(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		DRAW_PIXEL( 1, 16)
		HIGH_TILE()
		DRAW_PIXEL( 2,  0)
		DRAW_PIXEL( 3, 16)
		ADD_PITCH()
	}
}


static void draw_sprite8_3(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0, 12)
		HIGH_TILE()
		DRAW_PIXEL( 1,  0)
		DRAW_PIXEL( 2, 20)
		ADD_PITCH()
	}
}


static void draw_sprite8_2(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		HIGH_TILE()
		DRAW_PIXEL( 1,  0)
		ADD_PITCH()
	}
}


static void draw_sprite8_1(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		ADD_PITCH()
	}
}

/******************************************************************************/
/* draw sprite with flipx                                                     */
/******************************************************************************/

static void draw_sprite8_16_flipx(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL(15,  0)
		DRAW_PIXEL(14,  4)
		DRAW_PIXEL(13,  8)
		DRAW_PIXEL(12, 12)
		DRAW_PIXEL(11, 16)
		DRAW_PIXEL(10, 20)
		DRAW_PIXEL( 9, 24)
		DRAW_PIXEL( 8, 28)
		HIGH_TILE()
		DRAW_PIXEL( 7,  0)
		DRAW_PIXEL( 6,  4)
		DRAW_PIXEL( 5,  8)
		DRAW_PIXEL( 4, 12)
		DRAW_PIXEL( 3, 16)
		DRAW_PIXEL( 2, 20)
		DRAW_PIXEL( 1, 24)
		DRAW_PIXEL( 0, 28)
		ADD_PITCH()
	}
}


static void draw_sprite8_15_flipx(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL(14,  0)
		DRAW_PIXEL(13,  4)
		DRAW_PIXEL(12,  8)
		DRAW_PIXEL(11, 12)
		DRAW_PIXEL(10, 16)
		DRAW_PIXEL( 9, 20)
		DRAW_PIXEL( 8, 24)
		HIGH_TILE()
		DRAW_PIXEL( 7,  0)
		DRAW_PIXEL( 6,  4)
		DRAW_PIXEL( 5,  8)
		DRAW_PIXEL( 4, 12)
		DRAW_PIXEL( 3, 16)
		DRAW_PIXEL( 2, 20)
		DRAW_PIXEL( 1, 24)
		DRAW_PIXEL( 0, 28)
		ADD_PITCH()
	}
}


static void draw_sprite8_14_flipx(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL(13,  0)
		DRAW_PIXEL(12,  4)
		DRAW_PIXEL(11,  8)
		DRAW_PIXEL(10, 12)
		DRAW_PIXEL( 9, 16)
		DRAW_PIXEL( 8, 20)
		DRAW_PIXEL( 7, 24)
		HIGH_TILE()
		DRAW_PIXEL( 6,  0)
		DRAW_PIXEL( 5,  4)
		DRAW_PIXEL( 4,  8)
		DRAW_PIXEL( 3, 12)
		DRAW_PIXEL( 2, 16)
		DRAW_PIXEL( 1, 20)
		DRAW_PIXEL( 0, 24)
		ADD_PITCH()
	}
}


static void draw_sprite8_13_flipx(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL(12,  0)
		DRAW_PIXEL(11,  4)
		DRAW_PIXEL(10, 12)
		DRAW_PIXEL( 9, 16)
		DRAW_PIXEL( 8, 20)
		DRAW_PIXEL( 7, 24)
		HIGH_TILE()
		DRAW_PIXEL( 6,  0)
		DRAW_PIXEL( 5,  4)
		DRAW_PIXEL( 4,  8)
		DRAW_PIXEL( 3, 12)
		DRAW_PIXEL( 2, 20)
		DRAW_PIXEL( 1, 24)
		DRAW_PIXEL( 0, 28)
		ADD_PITCH()
	}
}


static void draw_sprite8_12_flipx(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL(11,  0)
		DRAW_PIXEL(10,  4)
		DRAW_PIXEL( 9,  8)
		DRAW_PIXEL( 8, 16)
		DRAW_PIXEL( 7, 20)
		DRAW_PIXEL( 6, 24)
		HIGH_TILE()
		DRAW_PIXEL( 5,  0)
		DRAW_PIXEL( 4,  4)
		DRAW_PIXEL( 3,  8)
		DRAW_PIXEL( 2, 16)
		DRAW_PIXEL( 1, 20)
		DRAW_PIXEL( 0, 24)
		ADD_PITCH()
	}
}


static void draw_sprite8_11_flipx(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL(10,  4)
		DRAW_PIXEL( 9,  8)
		DRAW_PIXEL( 8, 16)
		DRAW_PIXEL( 7, 20)
		DRAW_PIXEL( 6, 28)
		HIGH_TILE()
		DRAW_PIXEL( 5,  0)
		DRAW_PIXEL( 4,  4)
		DRAW_PIXEL( 3, 12)
		DRAW_PIXEL( 2, 16)
		DRAW_PIXEL( 1, 24)
		DRAW_PIXEL( 0, 28)
		ADD_PITCH()
	}
}


static void draw_sprite8_10_flipx(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 9,  0)
		DRAW_PIXEL( 8,  8)
		DRAW_PIXEL( 7, 16)
		DRAW_PIXEL( 6, 24)
		DRAW_PIXEL( 5, 28)
		HIGH_TILE()
		DRAW_PIXEL( 4,  0)
		DRAW_PIXEL( 3,  8)
		DRAW_PIXEL( 2, 16)
		DRAW_PIXEL( 1, 24)
		DRAW_PIXEL( 0, 28)
		ADD_PITCH()
	}
}


static void draw_sprite8_9_flipx(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 8,  0)
		DRAW_PIXEL( 7,  8)
		DRAW_PIXEL( 6, 16)
		DRAW_PIXEL( 5, 24)
		HIGH_TILE()
		DRAW_PIXEL( 4,  0)
		DRAW_PIXEL( 3,  8)
		DRAW_PIXEL( 2, 16)
		DRAW_PIXEL( 1, 24)
		DRAW_PIXEL( 0, 28)
		ADD_PITCH()
	}
}


static void draw_sprite8_8_flipx(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 7,  0)
		DRAW_PIXEL( 6,  8)
		DRAW_PIXEL( 5, 16)
		DRAW_PIXEL( 4, 24)
		HIGH_TILE()
		DRAW_PIXEL( 3,  0)
		DRAW_PIXEL( 2,  8)
		DRAW_PIXEL( 1, 16)
		DRAW_PIXEL( 0, 24)
		ADD_PITCH()
	}
}


static void draw_sprite8_7_flipx(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 6,  4)
		DRAW_PIXEL( 5, 12)
		DRAW_PIXEL( 4, 20)
		HIGH_TILE()
		DRAW_PIXEL( 3,  0)
		DRAW_PIXEL( 2,  8)
		DRAW_PIXEL( 1, 16)
		DRAW_PIXEL( 0, 24)
		ADD_PITCH()
	}
}


static void draw_sprite8_6_flipx(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 5,  4)
		DRAW_PIXEL( 4, 16)
		DRAW_PIXEL( 3, 28)
		HIGH_TILE()
		DRAW_PIXEL( 2,  0)
		DRAW_PIXEL( 1, 12)
		DRAW_PIXEL( 0, 24)
		ADD_PITCH()
	}
}


static void draw_sprite8_5_flipx(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 4,  8)
		DRAW_PIXEL( 3, 20)
		HIGH_TILE()
		DRAW_PIXEL( 2,  0)
		DRAW_PIXEL( 1, 12)
		DRAW_PIXEL( 0, 24)
		ADD_PITCH()
	}
}


static void draw_sprite8_4_flipx(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 3,  0)
		DRAW_PIXEL( 2, 16)
		HIGH_TILE()
		DRAW_PIXEL( 1,  0)
		DRAW_PIXEL( 0, 16)
		ADD_PITCH()
	}
}

static void draw_sprite8_3_flipx(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 2, 12)
		HIGH_TILE()
		DRAW_PIXEL( 1,  0)
		DRAW_PIXEL( 0, 20)
		ADD_PITCH()
	}
}

static void draw_sprite8_2_flipx(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		HIGH_TILE()
		DRAW_PIXEL( 1,  0)
		ADD_PITCH()
	}
}


static void draw_sprite8_1_flipx(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		HIGH_TILE()
		DRAW_PIXEL( 0,  0)
		ADD_PITCH()
	}
}


/******************************************************************************/
/* draw sprite no flipx                                                       */
/******************************************************************************/

static void draw_sprite16_16(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		DRAW_PIXEL( 1,  4)
		DRAW_PIXEL( 2,  8)
		DRAW_PIXEL( 3, 12)
		DRAW_PIXEL( 4, 16)
		DRAW_PIXEL( 5, 20)
		DRAW_PIXEL( 6, 24)
		DRAW_PIXEL( 7, 28)
		HIGH_TILE()
		DRAW_PIXEL( 8,  0)
		DRAW_PIXEL( 9,  4)
		DRAW_PIXEL(10,  8)
		DRAW_PIXEL(11, 12)
		DRAW_PIXEL(12, 16)
		DRAW_PIXEL(13, 20)
		DRAW_PIXEL(14, 24)
		DRAW_PIXEL(15, 28)
		ADD_PITCH()
	}
}


static void draw_sprite16_15(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		DRAW_PIXEL( 1,  4)
		DRAW_PIXEL( 2,  8)
		DRAW_PIXEL( 3, 12)
		DRAW_PIXEL( 4, 16)
		DRAW_PIXEL( 5, 20)
		DRAW_PIXEL( 6, 24)
		HIGH_TILE()
		DRAW_PIXEL( 7,  0)
		DRAW_PIXEL( 8,  4)
		DRAW_PIXEL( 9,  8)
		DRAW_PIXEL(10, 12)
		DRAW_PIXEL(11, 16)
		DRAW_PIXEL(12, 20)
		DRAW_PIXEL(13, 24)
		DRAW_PIXEL(14, 28)
		ADD_PITCH()
	}
}


static void draw_sprite16_14(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		DRAW_PIXEL( 1,  4)
		DRAW_PIXEL( 2,  8)
		DRAW_PIXEL( 3, 12)
		DRAW_PIXEL( 4, 16)
		DRAW_PIXEL( 5, 20)
		DRAW_PIXEL( 6, 24)
		HIGH_TILE()
		DRAW_PIXEL( 7,  0)
		DRAW_PIXEL( 8,  4)
		DRAW_PIXEL( 9,  8)
		DRAW_PIXEL(10, 12)
		DRAW_PIXEL(11, 16)
		DRAW_PIXEL(12, 20)
		DRAW_PIXEL(13, 24)
		ADD_PITCH()
	}
}


static void draw_sprite16_13(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		DRAW_PIXEL( 1,  4)
		DRAW_PIXEL( 2, 12)
		DRAW_PIXEL( 3, 16)
		DRAW_PIXEL( 4, 20)
		DRAW_PIXEL( 5, 24)
		HIGH_TILE()
		DRAW_PIXEL( 6,  0)
		DRAW_PIXEL( 7,  4)
		DRAW_PIXEL( 8,  8)
		DRAW_PIXEL( 9, 12)
		DRAW_PIXEL(10, 20)
		DRAW_PIXEL(11, 24)
		DRAW_PIXEL(12, 28)
		ADD_PITCH()
	}
}


static void draw_sprite16_12(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		DRAW_PIXEL( 1,  4)
		DRAW_PIXEL( 2,  8)
		DRAW_PIXEL( 3, 16)
		DRAW_PIXEL( 4, 20)
		DRAW_PIXEL( 5, 24)
		HIGH_TILE()
		DRAW_PIXEL( 6,  0)
		DRAW_PIXEL( 7,  4)
		DRAW_PIXEL( 8,  8)
		DRAW_PIXEL( 9, 16)
		DRAW_PIXEL(10, 20)
		DRAW_PIXEL(11, 24)
		ADD_PITCH()
	}
}


static void draw_sprite16_11(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  4)
		DRAW_PIXEL( 1,  8)
		DRAW_PIXEL( 2, 16)
		DRAW_PIXEL( 3, 20)
		DRAW_PIXEL( 4, 28)
		HIGH_TILE()
		DRAW_PIXEL( 5,  0)
		DRAW_PIXEL( 6,  4)
		DRAW_PIXEL( 7, 12)
		DRAW_PIXEL( 8, 16)
		DRAW_PIXEL( 9, 24)
		DRAW_PIXEL(10, 28)
		ADD_PITCH()
	}
}


static void draw_sprite16_10(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		DRAW_PIXEL( 1,  8)
		DRAW_PIXEL( 2, 16)
		DRAW_PIXEL( 3, 24)
		DRAW_PIXEL( 4, 28)
		HIGH_TILE()
		DRAW_PIXEL( 5,  0)
		DRAW_PIXEL( 6,  8)
		DRAW_PIXEL( 7, 16)
		DRAW_PIXEL( 8, 24)
		DRAW_PIXEL( 9, 28)
		ADD_PITCH()
	}
}


static void draw_sprite16_9(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		DRAW_PIXEL( 1,  8)
		DRAW_PIXEL( 2, 16)
		DRAW_PIXEL( 3, 24)
		HIGH_TILE()
		DRAW_PIXEL( 4,  0)
		DRAW_PIXEL( 5,  8)
		DRAW_PIXEL( 6, 16)
		DRAW_PIXEL( 7, 24)
		DRAW_PIXEL( 8, 28)
		ADD_PITCH()
	}
}


static void draw_sprite16_8(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		DRAW_PIXEL( 1,  8)
		DRAW_PIXEL( 2, 16)
		DRAW_PIXEL( 3, 24)
		HIGH_TILE()
		DRAW_PIXEL( 4,  0)
		DRAW_PIXEL( 5,  8)
		DRAW_PIXEL( 6, 16)
		DRAW_PIXEL( 7, 24)
		ADD_PITCH()
	}
}


static void draw_sprite16_7(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  4)
		DRAW_PIXEL( 1, 12)
		DRAW_PIXEL( 2, 20)
		HIGH_TILE()
		DRAW_PIXEL( 3,  0)
		DRAW_PIXEL( 4,  8)
		DRAW_PIXEL( 5, 16)
		DRAW_PIXEL( 6, 24)
		ADD_PITCH()
	}
}


static void draw_sprite16_6(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  4)
		DRAW_PIXEL( 1, 16)
		DRAW_PIXEL( 2, 28)
		HIGH_TILE()
		DRAW_PIXEL( 3,  0)
		DRAW_PIXEL( 4, 12)
		DRAW_PIXEL( 5, 24)
		ADD_PITCH()
	}
}


static void draw_sprite16_5(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  8)
		DRAW_PIXEL( 1, 20)
		HIGH_TILE()
		DRAW_PIXEL( 2,  0)
		DRAW_PIXEL( 3, 12)
		DRAW_PIXEL( 4, 24)
		ADD_PITCH()
	}
}


static void draw_sprite16_4(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		DRAW_PIXEL( 1, 16)
		HIGH_TILE()
		DRAW_PIXEL( 2,  0)
		DRAW_PIXEL( 3, 16)
		ADD_PITCH()
	}
}


static void draw_sprite16_3(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0, 12)
		HIGH_TILE()
		DRAW_PIXEL( 1,  0)
		DRAW_PIXEL( 2, 20)
		ADD_PITCH()
	}
}


static void draw_sprite16_2(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		HIGH_TILE()
		DRAW_PIXEL( 1,  0)
		ADD_PITCH()
	}
}


static void draw_sprite16_1(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		ADD_PITCH()
	}
}

/******************************************************************************/
/* draw sprite with flipx                                                     */
/******************************************************************************/

static void draw_sprite16_16_flipx(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL(15,  0)
		DRAW_PIXEL(14,  4)
		DRAW_PIXEL(13,  8)
		DRAW_PIXEL(12, 12)
		DRAW_PIXEL(11, 16)
		DRAW_PIXEL(10, 20)
		DRAW_PIXEL( 9, 24)
		DRAW_PIXEL( 8, 28)
		HIGH_TILE()
		DRAW_PIXEL( 7,  0)
		DRAW_PIXEL( 6,  4)
		DRAW_PIXEL( 5,  8)
		DRAW_PIXEL( 4, 12)
		DRAW_PIXEL( 3, 16)
		DRAW_PIXEL( 2, 20)
		DRAW_PIXEL( 1, 24)
		DRAW_PIXEL( 0, 28)
		ADD_PITCH()
	}
}


static void draw_sprite16_15_flipx(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL(14,  0)
		DRAW_PIXEL(13,  4)
		DRAW_PIXEL(12,  8)
		DRAW_PIXEL(11, 12)
		DRAW_PIXEL(10, 16)
		DRAW_PIXEL( 9, 20)
		DRAW_PIXEL( 8, 24)
		HIGH_TILE()
		DRAW_PIXEL( 7,  0)
		DRAW_PIXEL( 6,  4)
		DRAW_PIXEL( 5,  8)
		DRAW_PIXEL( 4, 12)
		DRAW_PIXEL( 3, 16)
		DRAW_PIXEL( 2, 20)
		DRAW_PIXEL( 1, 24)
		DRAW_PIXEL( 0, 28)
		ADD_PITCH()
	}
}


static void draw_sprite16_14_flipx(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL(13,  0)
		DRAW_PIXEL(12,  4)
		DRAW_PIXEL(11,  8)
		DRAW_PIXEL(10, 12)
		DRAW_PIXEL( 9, 16)
		DRAW_PIXEL( 8, 20)
		DRAW_PIXEL( 7, 24)
		HIGH_TILE()
		DRAW_PIXEL( 6,  0)
		DRAW_PIXEL( 5,  4)
		DRAW_PIXEL( 4,  8)
		DRAW_PIXEL( 3, 12)
		DRAW_PIXEL( 2, 16)
		DRAW_PIXEL( 1, 20)
		DRAW_PIXEL( 0, 24)
		ADD_PITCH()
	}
}


static void draw_sprite16_13_flipx(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL(12,  0)
		DRAW_PIXEL(11,  4)
		DRAW_PIXEL(10, 12)
		DRAW_PIXEL( 9, 16)
		DRAW_PIXEL( 8, 20)
		DRAW_PIXEL( 7, 24)
		HIGH_TILE()
		DRAW_PIXEL( 6,  0)
		DRAW_PIXEL( 5,  4)
		DRAW_PIXEL( 4,  8)
		DRAW_PIXEL( 3, 12)
		DRAW_PIXEL( 2, 20)
		DRAW_PIXEL( 1, 24)
		DRAW_PIXEL( 0, 28)
		ADD_PITCH()
	}
}


static void draw_sprite16_12_flipx(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL(11,  0)
		DRAW_PIXEL(10,  4)
		DRAW_PIXEL( 9,  8)
		DRAW_PIXEL( 8, 16)
		DRAW_PIXEL( 7, 20)
		DRAW_PIXEL( 6, 24)
		HIGH_TILE()
		DRAW_PIXEL( 5,  0)
		DRAW_PIXEL( 4,  4)
		DRAW_PIXEL( 3,  8)
		DRAW_PIXEL( 2, 16)
		DRAW_PIXEL( 1, 20)
		DRAW_PIXEL( 0, 24)
		ADD_PITCH()
	}
}


static void draw_sprite16_11_flipx(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL(10,  4)
		DRAW_PIXEL( 9,  8)
		DRAW_PIXEL( 8, 16)
		DRAW_PIXEL( 7, 20)
		DRAW_PIXEL( 6, 28)
		HIGH_TILE()
		DRAW_PIXEL( 5,  0)
		DRAW_PIXEL( 4,  4)
		DRAW_PIXEL( 3, 12)
		DRAW_PIXEL( 2, 16)
		DRAW_PIXEL( 1, 24)
		DRAW_PIXEL( 0, 28)
		ADD_PITCH()
	}
}


static void draw_sprite16_10_flipx(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 9,  0)
		DRAW_PIXEL( 8,  8)
		DRAW_PIXEL( 7, 16)
		DRAW_PIXEL( 6, 24)
		DRAW_PIXEL( 5, 28)
		HIGH_TILE()
		DRAW_PIXEL( 4,  0)
		DRAW_PIXEL( 3,  8)
		DRAW_PIXEL( 2, 16)
		DRAW_PIXEL( 1, 24)
		DRAW_PIXEL( 0, 28)
		ADD_PITCH()
	}
}


static void draw_sprite16_9_flipx(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 8,  0)
		DRAW_PIXEL( 7,  8)
		DRAW_PIXEL( 6, 16)
		DRAW_PIXEL( 5, 24)
		HIGH_TILE()
		DRAW_PIXEL( 4,  0)
		DRAW_PIXEL( 3,  8)
		DRAW_PIXEL( 2, 16)
		DRAW_PIXEL( 1, 24)
		DRAW_PIXEL( 0, 28)
		ADD_PITCH()
	}
}


static void draw_sprite16_8_flipx(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 7,  0)
		DRAW_PIXEL( 6,  8)
		DRAW_PIXEL( 5, 16)
		DRAW_PIXEL( 4, 24)
		HIGH_TILE()
		DRAW_PIXEL( 3,  0)
		DRAW_PIXEL( 2,  8)
		DRAW_PIXEL( 1, 16)
		DRAW_PIXEL( 0, 24)
		ADD_PITCH()
	}
}


static void draw_sprite16_7_flipx(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 6,  4)
		DRAW_PIXEL( 5, 12)
		DRAW_PIXEL( 4, 20)
		HIGH_TILE()
		DRAW_PIXEL( 3,  0)
		DRAW_PIXEL( 2,  8)
		DRAW_PIXEL( 1, 16)
		DRAW_PIXEL( 0, 24)
		ADD_PITCH()
	}
}


static void draw_sprite16_6_flipx(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 5,  4)
		DRAW_PIXEL( 4, 16)
		DRAW_PIXEL( 3, 28)
		HIGH_TILE()
		DRAW_PIXEL( 2,  0)
		DRAW_PIXEL( 1, 12)
		DRAW_PIXEL( 0, 24)
		ADD_PITCH()
	}
}


static void draw_sprite16_5_flipx(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 4,  8)
		DRAW_PIXEL( 3, 20)
		HIGH_TILE()
		DRAW_PIXEL( 2,  0)
		DRAW_PIXEL( 1, 12)
		DRAW_PIXEL( 0, 24)
		ADD_PITCH()
	}
}


static void draw_sprite16_4_flipx(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 3,  0)
		DRAW_PIXEL( 2, 16)
		HIGH_TILE()
		DRAW_PIXEL( 1,  0)
		DRAW_PIXEL( 0, 16)
		ADD_PITCH()
	}
}

static void draw_sprite16_3_flipx(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 2, 12)
		HIGH_TILE()
		DRAW_PIXEL( 1,  0)
		DRAW_PIXEL( 0, 20)
		ADD_PITCH()
	}
}

static void draw_sprite16_2_flipx(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		LOW_TILE()
		DRAW_PIXEL( 0,  0)
		HIGH_TILE()
		DRAW_PIXEL( 1,  0)
		ADD_PITCH()
	}
}


static void draw_sprite16_1_flipx(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy)
{
	UINT32 tile, pixel;
	int l = 0;

	while (zy--)
	{
		INCREMENT_LINE()
		HIGH_TILE()
		DRAW_PIXEL( 0,  0)
		ADD_PITCH()
	}
}

/******************************************************************************/

static void (*neo_draw_sprite8[2][16])(UINT32 *sprite, UINT8 *bitmap, UINT32 *paldata, int zy, int dy) =
{
	{
		draw_sprite8_1,  draw_sprite8_2,
		draw_sprite8_3,  draw_sprite8_4,
		draw_sprite8_5,  draw_sprite8_6,
		draw_sprite8_7,  draw_sprite8_8,
		draw_sprite8_9,  draw_sprite8_10,
		draw_sprite8_11, draw_sprite8_12,
		draw_sprite8_13, draw_sprite8_14,
		draw_sprite8_15, draw_sprite8_16
	},
	{
		draw_sprite8_1_flipx,  draw_sprite8_2_flipx,
		draw_sprite8_3_flipx,  draw_sprite8_4_flipx,
		draw_sprite8_5_flipx,  draw_sprite8_6_flipx,
		draw_sprite8_7_flipx,  draw_sprite8_8_flipx,
		draw_sprite8_9_flipx,  draw_sprite8_10_flipx,
		draw_sprite8_11_flipx, draw_sprite8_12_flipx,
		draw_sprite8_13_flipx, draw_sprite8_14_flipx,
		draw_sprite8_15_flipx, draw_sprite8_16_flipx
	}
};

static void (*neo_draw_sprite16[2][16])(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy) =
{
	{
		draw_sprite16_1,  draw_sprite16_2,
		draw_sprite16_3,  draw_sprite16_4,
		draw_sprite16_5,  draw_sprite16_6,
		draw_sprite16_7,  draw_sprite16_8,
		draw_sprite16_9,  draw_sprite16_10,
		draw_sprite16_11, draw_sprite16_12,
		draw_sprite16_13, draw_sprite16_14,
		draw_sprite16_15, draw_sprite16_16
	},
	{
		draw_sprite16_1_flipx,  draw_sprite16_2_flipx,
		draw_sprite16_3_flipx,  draw_sprite16_4_flipx,
		draw_sprite16_5_flipx,  draw_sprite16_6_flipx,
		draw_sprite16_7_flipx,  draw_sprite16_8_flipx,
		draw_sprite16_9_flipx,  draw_sprite16_10_flipx,
		draw_sprite16_11_flipx, draw_sprite16_12_flipx,
		draw_sprite16_13_flipx, draw_sprite16_14_flipx,
		draw_sprite16_15_flipx, draw_sprite16_16_flipx
	}
};


static void build_zoom_table(void)
{
	int i, my, zy, dday;

	for (zy = 0; zy <= 0xff; zy++)
	{
		dday = 0;

		for (my = 0; my < 0x21; my++)
		{
			if (zy != 0xff)
			{
				yskip[zy][my] = 0;
				dda_y_skip[zy][my][0] = 0;

				for(i = 0; i < 16; i++)
				{
					dda_y_skip[zy][my][i + 1] = 0;
					dday -= zy + 1;
					if (dday <= 0)
					{
						dday += 256;
						yskip[zy][my]++;
					}
					dda_y_skip[zy][my][yskip[zy][my]]++;
				}
			}
			else
			{
				dda_y_skip[zy][my][0] = 0;

				for(i = 1; i < 16; i++)
					dda_y_skip[zy][my][i] = 1;

				yskip[zy][my] = 16;
			}

			if (my < 0x10)
			{
				mmy[zy][my] = ((my * 16 * 256) / (zy + 1) + 15) / 16;
				if (mmy[zy][my] > 0x10)
					mmy[zy][my] = 0x10;
			}
		}
	}
}

/******************************************************************************/

void neogeo_vh_stop(void)
{
	if (neogeo_palettebank[0])
		free (neogeo_palettebank[0]);
	neogeo_palettebank[0] = NULL;

	if (neogeo_palettebank[1])
		free (neogeo_palettebank[1]);
	neogeo_palettebank[1] = NULL;

	if (neogeo_vidram16)
		free (neogeo_vidram16);
	neogeo_vidram16 = NULL;

	if (neogeo_ram16)
		free (neogeo_ram16);
	neogeo_ram16 = NULL;
}

static int common_vh_start(void)
{
	neogeo_palettebank[0] = NULL;
	neogeo_palettebank[1] = NULL;
	neogeo_vidram16 = NULL;

	neogeo_palettebank[0] = malloc(0x2000);
	if (!neogeo_palettebank[0])
	{
		neogeo_vh_stop();
		return 1;
	}
	memset(neogeo_palettebank[0],0,0x2000);

	neogeo_palettebank[1] = malloc(0x2000);
	if (!neogeo_palettebank[1])
	{
		neogeo_vh_stop();
		return 1;
	}
	memset(neogeo_palettebank[1],0,0x2000);

	/* 0x20000 bytes even though only 0x10c00 is used */
	neogeo_vidram16 = malloc(0x20000);
	if (!neogeo_vidram16)
	{
		neogeo_vh_stop();
		return 1;
	}
	memset(neogeo_vidram16,0,0x20000);

	neogeo_paletteram16 = neogeo_palettebank[0];
	neogeo_palette_index = 0;
	neogeo_vidram16_modulo = 1;
	neogeo_vidram16_offset = 0;
	fix_bank = 0;
	palette_swap_pending = 0;

	build_zoom_table();

	bitmap_pitch = Machine->scrbitmap->line[1] - Machine->scrbitmap->line[0];
	if (Machine->scrbitmap->depth != 8)
		bitmap_pitch /= 2;

	return 0;
}


static void decodetile(int tileno)
{
	unsigned char swap[128];
	UINT32 *gfxdata;
	int x,y;
	unsigned int pen;


	gfxdata = (UINT32 *)&memory_region(REGION_GFX2)[128 * tileno];

	memcpy(swap,gfxdata,128);

	for (y = 0; y < 16; y++)
	{
		UINT32 dw;

		dw = 0;
		for (x = 0; x < 8; x++)
		{
			pen  = ((swap[64 + 4*y + 3] >> x) & 1) << 3;
			pen |= ((swap[64 + 4*y + 1] >> x) & 1) << 2;
			pen |= ((swap[64 + 4*y + 2] >> x) & 1) << 1;
			pen |=  (swap[64 + 4*y    ] >> x) & 1;
			dw |= pen << 4*x;
			Machine->gfx[2]->pen_usage[tileno] |= (1 << pen);
		}
		*(gfxdata++) = dw;

		dw = 0;
		for (x = 0; x < 8; x++)
		{
			pen  = ((swap[4*y + 3] >> x) & 1) << 3;
			pen |= ((swap[4*y + 1] >> x) & 1) << 2;
			pen |= ((swap[4*y + 2] >> x) & 1) << 1;
			pen |=  (swap[4*y    ] >> x) & 1;
			dw |= pen << 4*x;
			Machine->gfx[2]->pen_usage[tileno] |= (1 << pen);
		}
		*(gfxdata++) = dw;
	}
}

int neogeo_mvs_vh_start(void)
{
	no_of_tiles=memory_region_length(REGION_GFX2)/128;
	if (no_of_tiles>0x10000) high_tile=1; else high_tile=0;
	if (no_of_tiles>0x20000) vhigh_tile=1; else vhigh_tile=0;
	if (no_of_tiles>0x40000) vvhigh_tile=1; else vvhigh_tile=0;
	Machine->gfx[2]->total_elements = no_of_tiles;
	if (Machine->gfx[2]->pen_usage)
		free(Machine->gfx[2]->pen_usage);
	Machine->gfx[2]->pen_usage = malloc(no_of_tiles * sizeof(int));
	memset(Machine->gfx[2]->pen_usage,0,no_of_tiles * sizeof(int));

	/* tiles are not decoded yet. They will be decoded later as they are used. */
	/* pen_usage is used as a marker of decoded tiles: if it is 0, then the tile */
	/* hasn't been decoded yet. */

	return common_vh_start();
}

/******************************************************************************/

static void swap_palettes(void)
{
	int i;

	for (i = 0; i < 0x2000 >> 1; i++)
	{
		data16_t newword = neogeo_paletteram16[i];
		int r,g,b;

		r = ((newword >> 4) & 0xf0) | ((newword >> 11) & 0x08);
		r |= (r >> 5);
		g = ( newword       & 0xf0) | ((newword >> 10) & 0x08);
		g |= (g >> 5);
		b = ((newword << 4) & 0xf0) | ((newword >>  9) & 0x08);
		b |= (b >> 5);

		palette_change_color(i, r, g, b);
	}

	palette_swap_pending = 0;
}

static void neogeo_setpalbank(int n)
{
	if (neogeo_palette_index != n)
	{
		neogeo_palette_index = n;
		neogeo_paletteram16 = neogeo_palettebank[n];
		palette_swap_pending = 1;
	}
}

WRITE16_HANDLER( neogeo_setpalbank0_16_w )
{
	neogeo_setpalbank(0);
}

WRITE16_HANDLER( neogeo_setpalbank1_16_w )
{
	neogeo_setpalbank(1);
}

READ16_HANDLER( neogeo_paletteram16_r )
{
	return neogeo_paletteram16[offset];
}

WRITE16_HANDLER( neogeo_paletteram16_w )
{
	data16_t oldword;
	int r,g,b;

	data &= 0x7fff;

	oldword = neogeo_paletteram16[offset];

	if (oldword == data)
		return;

	neogeo_paletteram16[offset] = data;

	r = ((data >> 4) & 0xf0) | ((data >> 11) & 0x08);
	r |= (r >> 5);
	g = ( data       & 0xf0) | ((data >> 10) & 0x08);
	g |= (g >> 5);
	b = ((data << 4) & 0xf0) | ((data >>  9) & 0x08);
	b |= (b >> 5);

	palette_change_color(offset, r, g, b);
}

/******************************************************************************/

static const UINT8 *neogeo_palette(const struct rectangle *clip)
{
	int color, code, pal_base;
	int colmask[256];
	int sx = 0, sy = 0, oy = 0, my = 0, zx = 0, zy = 0;
	int offs, count, y, i;
	int tileno, tileatr, t1, t3;
	char fullmode = 0;
	unsigned int *pen_usage; /* Save some struct derefs */
	struct GfxElement *gfx = Machine->gfx[2]; /* Save constant struct dereference */

	if (Machine->scrbitmap->depth == 16 || !palette_used_colors)
	{
		return palette_recalc();
	}

	palette_init_used_colors();

	/* character foreground */
	pen_usage = Machine->gfx[fix_bank]->pen_usage;
	pal_base = Machine->drv->gfxdecodeinfo[fix_bank].color_codes_start;
	for (color = 0; color < 16; color++)
		colmask[color] = 0;
	for (offs = 0xe000 >> 1; offs < 0xea00 >> 1; offs++)
	{
		code = neogeo_vidram16[offs];
		color = code >> 12;
		colmask[color] |= pen_usage[code&0xfff];
	}
	for (color = 0; color < 16; color++)
	{
		for (i = 1;i < 16;i++)
		{
			if (colmask[color] & (1 << i))
				palette_used_colors[pal_base + 16 * color + i] = PALETTE_COLOR_VISIBLE;
		}
	}

	/* Tiles */
	pen_usage= Machine->gfx[2]->pen_usage;
	pal_base = Machine->drv->gfxdecodeinfo[2].color_codes_start;
	for (color = 0; color < 256; color++)
		colmask[color] = 0;
	for (count = 0; count < 0x300 >> 1; count++)
	{
		t3 = neogeo_vidram16[(0x10000 >> 1) + count];
		t1 = neogeo_vidram16[(0x10400 >> 1) + count];

		/* If this bit is set this new column is placed next to last one */
		if (t1 & 0x40)
		{
			/* No point doing anything if tile strip is 0 */
			if (my == 0) continue;

			sx += zx;
			/* Get new zoom for this column */
			zx = (t3 >> 8) & 0x0f;
			if (gururin_fix && (t3 == 0 || t3 == 0x147f))
				zx = 0x0f;
			zx++;
			sy = oy;
		}
		else /* nope it is a new block */
		{
			/* Number of tiles in this strip */
			my = t1 & 0x3f;
			if (my == 0)
				continue; /* No point doing anything if tile strip is 0 */
			else if (my == 0x20)
				fullmode = 1;
			else if (my >= 0x21)
				fullmode = 2;	/* most games use 0x21, but  Alpha Mission II uses 0x3f */
			else
				fullmode = 0;

			/* Sprite scaling */
			zy = t3 & 0xff;
			/* Get new X zoom for this column */
			zx = (t3 >> 8) & 0x0f;
			if (gururin_fix && (t3 == 0 || t3 == 0x147f))
			{
				zx = 0x0f;
				zy = 0xff;
			}
			if (zy == 0)
			{
				/* No point doing anything if tile Y zoom is 0 */
				my = 0;
				continue;
			}
			zx++;

			sy = 0x200 - (t1 >> 7);

			if (clip->max_y - clip->min_y > 8 ||	/* kludge to improve the ssideki games */
				clip->min_y == Machine->visible_area.min_y)
			{
				if (sy > 0x110) sy -= 0x200;
				if (fullmode == 2 || (fullmode == 1 && zy == 0xff))
				{
					while (sy < 0) sy += 2 * (zy + 1);
				}
			}
			oy = sy;

			if (zy < 0xff && my < 0x10)
				my = mmy[zy][my];
			else if (my > 0x20)
				my = 0x20;

			sx = (neogeo_vidram16[(0x10800 >> 1) + count] >> 7);
		}

		if (sx >= 320)
			sx -= 0x200;

		if (sx + zx <= 8)
			continue;

		offs = count << 6;

		/* my holds the number of tiles in each vertical multisprite block */
		for (y = 0; y < my ; y++)
		{
			if (fullmode == 2 || (fullmode == 1 && zy == 0xff))
			{
				if (sy >= 248) sy -= 2 * (zy + 1);
			}
			else if (fullmode == 1)
			{
				if (y == 0x10) sy -= 2 * (zy + 1);
			}
			else if (sy > 0x110)
			{
				sy -= 0x200;	/* NS990105 mslug2 fix */
			}

			if (sy + yskip[zy][y] > clip->min_y && sy <= clip->max_y)
			{
				tileno  = neogeo_vidram16[offs++];
				tileatr = neogeo_vidram16[offs++];

				if (high_tile && tileatr & 0x10)   tileno += 0x10000;
				if (vhigh_tile && tileatr & 0x20)  tileno += 0x20000;
				if (vvhigh_tile && tileatr & 0x40) tileno += 0x40000;

				if (tileatr & 0x08)
				{
					tileno = (tileno & ~7) + (neogeo_frame_counter & 7);
				}
				else if (tileatr & 0x04)
				{
					tileno = (tileno & ~3) + (neogeo_frame_counter & 3);
				}

				l_y_skip = dda_y_skip[zy][y];

				/* Safety feature */
				tileno = tileno % no_of_tiles;

				if (!gfx->pen_usage[tileno])	/* decode tile if it hasn't been yet */
					decodetile(tileno);

				/* Check for total transparency, no need to draw */
				if (gfx->pen_usage[tileno] & ~1)
					colmask[tileatr >> 8] |= pen_usage[tileno];
			}
			else offs += 2;

			sy += yskip[zy][y];
		}  /* for y */
	}  /* for count */


	for (color = 0; color < 256; color++)
	{
		for (i = 1;i < 16;i++)
		{
			if (colmask[color] & (1 << i))
				palette_used_colors[pal_base + 16 * color + i] = PALETTE_COLOR_VISIBLE;
		}
	}

	palette_used_colors[4095] = PALETTE_COLOR_VISIBLE;

	return palette_recalc();
}

/******************************************************************************/

WRITE16_HANDLER( neogeo_vidram16_offset_w )
{
	COMBINE_DATA(&neogeo_vidram16_offset);
}

READ16_HANDLER( neogeo_vidram16_data_r )
{
	return neogeo_vidram16[neogeo_vidram16_offset];
}

WRITE16_HANDLER( neogeo_vidram16_data_w )
{
	COMBINE_DATA(&neogeo_vidram16[neogeo_vidram16_offset]);
	neogeo_vidram16_offset += neogeo_vidram16_modulo;
	neogeo_vidram16_offset &= 0xffff;
}

/* Modulo can become negative , Puzzle Bobble Super Sidekicks and a lot */
/* of other games use this */
WRITE16_HANDLER( neogeo_vidram16_modulo_w )
{
	COMBINE_DATA(&neogeo_vidram16_modulo);
}

READ16_HANDLER( neogeo_vidram16_modulo_r )
{
	return neogeo_vidram16_modulo;
}


WRITE16_HANDLER( neo_board_fix_16_w )
{
	fix_bank = 1;
}

WRITE16_HANDLER( neo_game_fix_16_w )
{
	fix_bank = 0;
}

/******************************************************************************/

INLINE void neo_draw_gfx(unsigned char **line, UINT32 *paldata,
		unsigned int code, int flipx, int flipy, int sx, int sy, int zx, int zy,
		const struct rectangle *clip, int depth)
{
	int oy, ey, dy;
	UINT32 *fspr;

	oy = sy;
	ey = sy + zy -1;

	if (sy < clip->min_y) sy = clip->min_y;
	if (ey >= clip->max_y) ey = clip->max_y;

	fspr = (UINT32 *)memory_region(REGION_GFX2);

	if (!flipy) /* normal */
	{
		dy = 2;
		fspr += code * 32 + (sy - oy) * 2;
	}
	else /* Y flip */
	{
		dy = -2;
		fspr += (code + 1) * 32 - 2 - (sy - oy) * 2;
	}

	zy = ey - sy + 1;

	if (depth == 8)
		(*neo_draw_sprite8[flipx][zx - 1])(fspr, &line[sy][sx], paldata, zy, dy);
	else
		(*neo_draw_sprite16[flipx][zx - 1])(fspr, (UINT16 *)&line[sy][sx * 2], paldata, zy, dy);
}


void NeoMVSDrawGfx(unsigned char **line,const struct GfxElement *gfx,
		unsigned int code,unsigned int color,int flipx,int flipy,int sx,int sy,
		int zx,int zy,const struct rectangle *clip)
{
	/* Safety feature */
	code = code % no_of_tiles;

	if (!gfx->pen_usage[code])	/* decode tile if it hasn't been yet */
		decodetile(code);

	/* Check for total transparency, no need to draw */
	if (gfx->pen_usage[code] & ~1)
	{
		neo_draw_gfx(line,
			&gfx->colortable[gfx->color_granularity * color],
			code,
			flipx, flipy,
			sx,sy,zx,zy,
			clip,
			8
		);
	}
}

void NeoMVSDrawGfx16(unsigned char **line,const struct GfxElement *gfx,
		unsigned int code,unsigned int color,int flipx,int flipy,int sx,int sy,
		int zx,int zy,const struct rectangle *clip)
{
	/* Safety feature */
	code = code % no_of_tiles;

	if (!gfx->pen_usage[code])	/* decode tile if it hasn't been yet */
		decodetile(code);

	/* Check for total transparency, no need to draw */
	if (gfx->pen_usage[code] & ~1)
	{
		neo_draw_gfx(line,
			&gfx->colortable[gfx->color_granularity * color],
			code,
			flipx, flipy,
			sx,sy,zx,zy,
			clip,
			16
		);
	}
}


/* This fix Japanese font in test mode. */
INLINE int convert_jfont(int code)
{
	if (sdodgeb_fix)
	{
		switch (code)
		{
		case 0x016e:
		case 0x0166:
		case 0x0173:
			code -= 0xc0;
			break;

		case 0x026e:
		case 0x0273:
		case 0x0266:
			code -= 0x180;
			break;

		case 0x010d:
			code = 0x008d;
			break;
		case 0x020d:
			code = 0x00cd;
			break;
		}
	}

	if (code >= 0x100 && code <= 0x2ff)
		code += 0x200;

	return code;
}


/******************************************************************************/

static void screenrefresh(struct osd_bitmap *bitmap, const struct rectangle *clip)
{
	int sx = 0, sy = 0, oy = 0, my = 0, zx = 0, zy = 0;
	int offs, count, y, x;
	int tileno, tileatr, t1, t3;
	char fullmode = 0;
	unsigned char **line = bitmap->line;
	unsigned int *pen_usage;
	struct GfxElement *gfx = Machine->gfx[2]; /* Save constant struct dereference */

	get_sf_keys();

	/* Palette swap occured after last frame but before this one */
	if (palette_swap_pending) swap_palettes();

	/* Do compressed palette stuff */
	neogeo_palette(clip);
	/* no need to check the return code since we redraw everything each frame */

	if ((sf_always || paused) && alt_bg) {
   	    palette_change_color(4094,255,0,255);
            fillbitmap(bitmap,Machine->pens[4094],clip);
	}
    
    	else {
    		fillbitmap(bitmap,Machine->pens[4095],clip);
    	}

	/* Draw sprites */
	for (count = 0; count < 0x300 >> 1; count++)
	{
		t3 = neogeo_vidram16[(0x10000 >> 1) + count];
		t1 = neogeo_vidram16[(0x10400 >> 1) + count];

        if ((sf_always || paused) && (!sprlayer_on || (!sprites_on && !(count >= tile_start && count <= tile_end)))) {
            continue;
        }

		/* If this bit is set this new column is placed next to last one */
		if (t1 & 0x40)
		{
			/* No point doing anything if tile strip is 0 */
			if (my == 0) continue;

			sx += zx;
			/* Get new zoom for this column */
			zx = (t3 >> 8) & 0x0f;
			if (gururin_fix && (t3 == 0 || t3 == 0x147f))
				zx = 0x0f;
			zx++;
			sy = oy;
		}
		else /* nope it is a new block */
		{
			/* Number of tiles in this strip */
			my = t1 & 0x3f;
			if (my == 0)
				continue; /* No point doing anything if tile strip is 0 */
			else if (my == 0x20)
				fullmode = 1;
			else if (my >= 0x21)
				fullmode = 2;	/* most games use 0x21, but  Alpha Mission II uses 0x3f */
			else
				fullmode = 0;

			/* Sprite scaling */
			zy = t3 & 0xff;
			/* Get new X zoom for this column */
			zx = (t3 >> 8) & 0x0f;
			if (gururin_fix && (t3 == 0 || t3 == 0x147f))
			{
				zx = 0x0f;
				zy = 0xff;
			}
			if (zy == 0)
			{
				/* No point doing anything if tile Y zoom is 0 */
				my = 0;
				continue;
			}
			zx++;

			sy = 0x200 - (t1 >> 7);

			if (clip->max_y - clip->min_y > 8 ||	/* kludge to improve the ssideki games */
				clip->min_y == Machine->visible_area.min_y)
			{
				if (sy > 0x110) sy -= 0x200;
				if (fullmode == 2 || (fullmode == 1 && zy == 0xff))
				{
					while (sy < 0) sy += 2 * (zy + 1);
				}
			}
			oy = sy;

			if (zy < 0xff && my < 0x10)
				my = mmy[zy][my];
			else if (my > 0x20)
				my = 0x20;

			sx = (neogeo_vidram16[(0x10800 >> 1) + count] >> 7);
		}

		if (sx >= 320)
			sx -= 0x200;

		if (sx + zx <= 8)
			continue;

		offs = count << 6;

		/* my holds the number of tiles in each vertical multisprite block */
		for (y = 0; y < my ; y++)
		{
			if (fullmode == 2 || (fullmode == 1 && zy == 0xff))
			{
				if (sy >= 248) sy -= 2 * (zy + 1);
			}
			else if (fullmode == 1)
			{
				if (y == 0x10) sy -= 2 * (zy + 1);
			}
			else if (sy > 0x110)
			{
				sy -= 0x200;	/* NS990105 mslug2 fix */
			}

			if (sy + yskip[zy][y] > clip->min_y && sy <= clip->max_y)
			{
				tileno  = neogeo_vidram16[offs++];
				tileatr = neogeo_vidram16[offs++];

				if (high_tile && tileatr & 0x10)   tileno += 0x10000;
				if (vhigh_tile && tileatr & 0x20)  tileno += 0x20000;
				if (vvhigh_tile && tileatr & 0x40) tileno += 0x40000;

				if (tileatr & 0x08)
				{
					tileno = (tileno & ~7) + (neogeo_frame_counter & 7);
				}
				else if (tileatr & 0x04)
				{
					tileno = (tileno & ~3) + (neogeo_frame_counter & 3);
				}

				l_y_skip = dda_y_skip[zy][y];

				/* Safety feature */
				tileno = tileno % no_of_tiles;

				if (!gfx->pen_usage[tileno])	/* decode tile if it hasn't been yet */
					decodetile(tileno);

				/* Check for total transparency, no need to draw */
				if (gfx->pen_usage[tileno] & ~1)
				{
					UINT32 *paldata = &gfx->colortable[gfx->color_granularity * (tileatr >> 8)];

					neo_draw_gfx(line,
						paldata,
						tileno,
						tileatr & 0x01,tileatr & 0x02,
						sx,sy,zx,yskip[zy][y],
						clip,
						bitmap->depth
					);
				}
			}
			else offs += 2;

			sy += yskip[zy][y];
		}  /* for y */
	        ng_tile_counter=count;
	}  /* for count */


	/* Save some struct de-refs */
	gfx = Machine->gfx[fix_bank];
	pen_usage=gfx->pen_usage;

	/* Character foreground */
	for (y=clip->min_y / 8; y <= clip->max_y / 8; y++)
	{
		for (x = 0; x < 40; x++)
		{

			int byte1 = neogeo_vidram16[(0xe000 >> 1) + y + 32 * x];
			int byte2 = byte1 >> 12;
			byte1 = byte1 & 0xfff;

			if (fix_bank && !neogeo_machine_country)
				byte1 = convert_jfont(byte1);

			if ((pen_usage[byte1] & ~1) == 0) continue;

	                if ((sf_always || paused) && !fglayer_on) {
		        }
		            
		        else {

			drawgfx(bitmap,gfx,
					byte1,
					byte2,
					0,0,
					x*8,y*8,
					clip,TRANSPARENCY_PEN,0);
			}
		}
	}
}

void neogeo_vh_screenrefresh(struct osd_bitmap *bitmap,int full_refresh)
{
	screenrefresh(bitmap,&Machine->visible_area);
	
	if (info_on) {
        	display_status(bitmap);
    	}

}

static int next_update_first_line;
extern int neogeo_raster_enable;

void neogeo_vh_raster_partial_refresh(struct osd_bitmap *bitmap,int current_line)
{
	struct rectangle clip;

	if (current_line < next_update_first_line)
		next_update_first_line = 0;

	clip.min_x = Machine->visible_area.min_x;
	clip.max_x = Machine->visible_area.max_x;
	clip.min_y = next_update_first_line;
	clip.max_y = current_line;
	if (clip.min_y < Machine->visible_area.min_y)
		clip.min_y = Machine->visible_area.min_y;
	if (clip.max_y > Machine->visible_area.max_y)
		clip.max_y = Machine->visible_area.max_y;

	if (clip.max_y >= clip.min_y)
	{
//logerror("refresh %d-%d\n",clip.min_y,clip.max_y);
		screenrefresh(bitmap,&clip);

		if (neogeo_raster_enable == 2)
		{
			int x;
			for (x = clip.min_x;x <= clip.max_x;x++)
			{
				if (x & 8)
					plot_pixel(bitmap,x,clip.max_y,Machine->uifont->colortable[1]);
			}
		}
	}

	next_update_first_line = current_line + 1;

}

void neogeo_vh_raster_screenrefresh(struct osd_bitmap *bitmap,int full_refresh)
{

    get_sf_keys();
    
    if (!sf_raster) {
        	screenrefresh(bitmap,&Machine->visible_area);
    }
        
    if (info_on) {
        display_status(bitmap);
    }
}
